@@ -11,15 +11,14 @@ from inspect import isclass
 from inspect import isfunction
 import sys
 
-import six
-
 from ddtrace import Pin
 from ddtrace import config
-from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
 from ddtrace.constants import SPAN_MEASURED_KEY
 from ddtrace.contrib import dbapi
 from ddtrace.contrib import func_name
 
+from ...internal.utils import get_argument_value
+
 
 try:
     from psycopg2._psycopg import cursor as psycopg_cursor_cls
@@ -34,17 +33,12 @@ from ddtrace.ext import http
 from ddtrace.ext import sql as sqlx
 from ddtrace.internal.compat import maybe_stringify
 from ddtrace.internal.logger import get_logger
-from ddtrace.propagation.utils import from_wsgi_header
-from ddtrace.utils.formats import asbool
-from ddtrace.utils.formats import get_env
-from ddtrace.vendor import debtcollector
+from ddtrace.internal.utils.formats import asbool
+from ddtrace.internal.utils.formats import get_env
 from ddtrace.vendor import wrapt
 
-from . import conf
 from . import utils
 from .. import trace_utils
-from .compat import get_resolver
-from .compat import user_is_authenticated
 
 
 log = get_logger(__name__)
@@ -59,8 +53,8 @@ config._add(
         trace_fetch_methods=asbool(get_env("django", "trace_fetch_methods", default=False)),
         distributed_tracing_enabled=True,
         instrument_middleware=asbool(get_env("django", "instrument_middleware", default=True)),
-        instrument_databases=True,
-        instrument_caches=True,
+        instrument_databases=asbool(get_env("django", "instrument_databases", default=True)),
+        instrument_caches=asbool(get_env("django", "instrument_caches", default=True)),
         analytics_enabled=None,  # None allows the value to be overridden by the global config
         analytics_sample_rate=None,
         trace_query_string=None,  # Default to global config
@@ -71,11 +65,6 @@ config._add(
 )
 
 
-# Set on patch, when django is imported
-Resolver404 = None
-DJANGO22 = None
-
-
 def patch_conn(django, conn):
     def cursor(django, pin, func, instance, args, kwargs):
         alias = getattr(conn, "alias", "default")
@@ -122,30 +111,6 @@ def instrument_dbs(django):
         )
 
 
-def _set_request_tags(django, span, request):
-    span._set_str_tag("django.request.class", func_name(request))
-
-    user = getattr(request, "user", None)
-    if user is not None:
-        # Note: getattr calls to user / user_is_authenticated may result in ImproperlyConfigured exceptions from
-        # Django's get_user_model():
-        # https://github.com/django/django/blob/a464ead29db8bf6a27a5291cad9eb3f0f3f0472b/django/contrib/auth/__init__.py
-        try:
-            if hasattr(user, "is_authenticated"):
-                span._set_str_tag("django.user.is_authenticated", str(user_is_authenticated(user)))
-
-            uid = getattr(user, "pk", None)
-            if uid:
-                span._set_str_tag("django.user.id", str(uid))
-
-            if config.django.include_user_name:
-                username = getattr(user, "username", None)
-                if username:
-                    span._set_str_tag("django.user.name", username)
-        except Exception:
-            log.debug("Error retrieving authentication information for user %r", user, exc_info=True)
-
-
 @trace_utils.with_traced_module
 def traced_cache(django, pin, func, instance, args, kwargs):
     if not config.django.instrument_caches:
@@ -209,10 +174,6 @@ def traced_populate(django, pin, func, instance, args, kwargs):
 
     settings = django.conf.settings
 
-    if hasattr(settings, "DATADOG_TRACE"):
-        debtcollector.deprecate(("Using DATADOG_TRACE Django settings are no longer supported. "))
-        conf.configure_from_settings(pin, config.django, settings.DATADOG_TRACE)
-
     # Instrument databases
     if config.django.instrument_databases:
         try:
@@ -291,19 +252,28 @@ def traced_load_middleware(django, pin, func, instance, args, kwargs):
             base = ".".join(split[:-1])
             attr = split[-1]
 
-            # Function-based middleware is a factory which returns a handler function for requests.
-            # So instead of tracing the factory, we want to trace its returned value.
-            # We wrap the factory to return a traced version of the handler function.
-            def wrapped_factory(func, instance, args, kwargs):
-                # r is the middleware handler function returned from the factory
-                r = func(*args, **kwargs)
-                if r:
-                    return wrapt.FunctionWrapper(r, traced_func(django, "django.middleware", resource=mw_path))
-                # If r is an empty middleware function (i.e. returns None), don't wrap since NoneType cannot be called
-                else:
-                    return r
+            # DEV: We need to have a closure over `mw_path` for the resource name or else
+            # all function based middleware will share the same resource name
+            def _wrapper(resource):
+                # Function-based middleware is a factory which returns a handler function for requests.
+                # So instead of tracing the factory, we want to trace its returned value.
+                # We wrap the factory to return a traced version of the handler function.
+                def wrapped_factory(func, instance, args, kwargs):
+                    # r is the middleware handler function returned from the factory
+                    r = func(*args, **kwargs)
+                    if r:
+                        return wrapt.FunctionWrapper(
+                            r,
+                            traced_func(django, "django.middleware", resource=resource),
+                        )
+                    # If r is an empty middleware function (i.e. returns None), don't wrap since
+                    # NoneType cannot be called
+                    else:
+                        return r
 
-            trace_utils.wrap(base, attr, wrapped_factory)
+                return wrapped_factory
+
+            trace_utils.wrap(base, attr, _wrapper(resource=mw_path))
 
         # Instrument class-based middleware
         elif isclass(mw):
@@ -338,7 +308,7 @@ def traced_get_response(django, pin, func, instance, args, kwargs):
     This method invokes the middleware chain and returns the response generated by the chain.
     """
 
-    request = kwargs.get("request", args[0])
+    request = get_argument_value(args, kwargs, 0, "request")
     if request is None:
         return func(*args, **kwargs)
 
@@ -346,122 +316,20 @@ def traced_get_response(django, pin, func, instance, args, kwargs):
 
     with pin.tracer.trace(
         "django.request",
-        resource=request.method,
+        resource=utils.REQUEST_DEFAULT_RESOURCE,
         service=trace_utils.int_service(pin, config.django),
         span_type=SpanTypes.WEB,
     ) as span:
+        utils._before_request_tags(pin, span, request)
         span.metrics[SPAN_MEASURED_KEY] = 1
 
-        analytics_sr = config.django.get_analytics_sample_rate(use_global_config=True)
-        if analytics_sr is not None:
-            span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, analytics_sr)
-
-        # Set HTTP Request tags
-        response = func(*args, **kwargs)
-
+        response = None
         try:
-            # Get resolver match result and build resource name pieces
-            resolver_match = request.resolver_match
-            if not resolver_match:
-                # The request quite likely failed (e.g. 404) so we do the resolution anyway.
-                resolver = get_resolver(getattr(request, "urlconf", None))
-                resolver_match = resolver.resolve(request.path_info)
-            handler = func_name(resolver_match[0])
-
-            if config.django.use_handler_resource_format:
-                span.resource = " ".join((span.resource, handler))
-            elif config.django.use_legacy_resource_format:
-                span.resource = handler
-            else:
-                # In Django >= 2.2.0 we can access the original route or regex pattern
-                # TODO: Validate if `resolver.pattern.regex.pattern` is available on django<2.2
-                if DJANGO22:
-                    # Determine the resolver and resource name for this request
-                    route = utils.get_django_2_route(request, resolver_match)
-                    if route:
-                        span.resource = " ".join((request.method, route))
-                        span._set_str_tag("http.route", route)
-                else:
-                    span.resource = " ".join((request.method, handler))
-
-            span._set_str_tag("django.view", resolver_match.view_name)
-            utils.set_tag_array(span, "django.namespace", resolver_match.namespaces)
-
-            # Django >= 2.0.0
-            if hasattr(resolver_match, "app_names"):
-                utils.set_tag_array(span, "django.app", resolver_match.app_names)
-
-        except Resolver404:
-            # Normalize all 404 requests into a single resource name
-            # DEV: This is for potential cardinality issues
-            span.resource = " ".join((request.method, "404"))
-        except Exception:
-            log.debug(
-                "Failed to resolve request path %r with path info %r",
-                request,
-                getattr(request, "path_info", "not-set"),
-                exc_info=True,
-            )
-
-        # Note: this call must be done after the function call because
-        # some attributes (like `user`) are added to the request through
-        # the middleware chain
-        _set_request_tags(django, span, request)
-
-        if response:
-            status = response.status_code
-            span._set_str_tag("django.response.class", func_name(response))
-            if hasattr(response, "template_name"):
-                # template_name is a bit of a misnomer, as it could be any of:
-                # a list of strings, a tuple of strings, a single string, or an instance of Template
-                # for more detail, see:
-                # https://docs.djangoproject.com/en/3.0/ref/template-response/#django.template.response.SimpleTemplateResponse.template_name
-                template = response.template_name
-
-                if isinstance(template, six.string_types):
-                    template_names = [template]
-                elif isinstance(
-                    template,
-                    (
-                        list,
-                        tuple,
-                    ),
-                ):
-                    template_names = template
-                elif hasattr(template, "template"):
-                    # ^ checking by attribute here because
-                    # django backend implementations don't have a common base
-                    # `.template` is also the most consistent across django versions
-                    template_names = [template.template.name]
-                else:
-                    template_names = None
-
-                utils.set_tag_array(span, "django.response.template", template_names)
-
-            url = utils.get_request_uri(request)
-
-            if DJANGO22:
-                request_headers = request.headers
-            else:
-                request_headers = {}
-                for header, value in request.META.items():
-                    name = from_wsgi_header(header)
-                    if name:
-                        request_headers[name] = value
-
-            response_headers = dict(response.items())
-            trace_utils.set_http_meta(
-                span,
-                config.django,
-                method=request.method,
-                url=url,
-                status_code=status,
-                query=request.META.get("QUERY_STRING", None),
-                request_headers=request_headers,
-                response_headers=response_headers,
-            )
-
-        return response
+            response = func(*args, **kwargs)
+            return response
+        finally:
+            # DEV: Always set these tags, this is where `span.resource` is set
+            utils._after_request_tags(pin, span, request, response)
 
 
 @trace_utils.with_traced_module
@@ -567,7 +435,17 @@ def traced_as_view(django, pin, func, instance, args, kwargs):
     except Exception:
         log.debug("Failed to instrument Django view %r", instance, exc_info=True)
     view = func(*args, **kwargs)
-    return wrapt.FunctionWrapper(view, traced_func(django, "django.view", resource=func_name(view)))
+    return wrapt.FunctionWrapper(view, traced_func(django, "django.view", resource=func_name(instance)))
+
+
+@trace_utils.with_traced_module
+def traced_get_asgi_application(django, pin, func, instance, args, kwargs):
+    from ddtrace.contrib.asgi import TraceMiddleware
+
+    def django_asgi_modifier(span, scope):
+        span.name = "django.request"
+
+    return TraceMiddleware(func(*args, **kwargs), integration_config=config.django, span_modifier=django_asgi_modifier)
 
 
 def _patch(django):
@@ -582,6 +460,21 @@ def _patch(django):
         trace_utils.wrap(django, "core.handlers.base.BaseHandler.load_middleware", traced_load_middleware(django))
 
     trace_utils.wrap(django, "core.handlers.base.BaseHandler.get_response", traced_get_response(django))
+    if hasattr(django.core.handlers.base.BaseHandler, "get_response_async"):
+        # Have to inline this import as the module contains syntax incompatible with Python 3.5 and below
+        from ._asgi import traced_get_response_async
+
+        trace_utils.wrap(django, "core.handlers.base.BaseHandler.get_response_async", traced_get_response_async(django))
+
+        # Only wrap get_asgi_application if get_response_async exists. Otherwise we will effectively double-patch
+        # because get_response and get_asgi_application will be used.
+        if "django.core.asgi" not in sys.modules:
+            try:
+                import django.core.asgi
+            except ImportError:
+                pass
+            else:
+                trace_utils.wrap(django, "core.asgi.get_asgi_application", traced_get_asgi_application(django))
 
     # DEV: this check will be replaced with import hooks in the future
     if "django.template.base" not in sys.modules:
@@ -591,7 +484,10 @@ def _patch(django):
     # DEV: this check will be replaced with import hooks in the future
     if "django.conf.urls.static" not in sys.modules:
         import django.conf.urls.static
-    trace_utils.wrap(django, "conf.urls.url", traced_urls_path(django))
+
+    if django.VERSION < (4, 0, 0):
+        trace_utils.wrap(django, "conf.urls.url", traced_urls_path(django))
+
     if django.VERSION >= (2, 0, 0):
         trace_utils.wrap(django, "urls.path", traced_urls_path(django))
         trace_utils.wrap(django, "urls.re_path", traced_urls_path(django))
@@ -603,17 +499,15 @@ def _patch(django):
 
 
 def patch():
-    global Resolver404, DJANGO22
-
     # DEV: this import will eventually be replaced with the module given from an import hook
     import django
 
     if django.VERSION < (1, 10, 0):
-        Resolver404 = django.core.urlresolvers.Resolver404
+        utils.Resolver404 = django.core.urlresolvers.Resolver404
     else:
-        Resolver404 = django.urls.exceptions.Resolver404
+        utils.Resolver404 = django.urls.exceptions.Resolver404
 
-    DJANGO22 = django.VERSION >= (2, 2, 0)
+    utils.DJANGO22 = django.VERSION >= (2, 2, 0)
 
     if getattr(django, "_datadog_patch", False):
         return
@@ -626,6 +520,7 @@ def _unpatch(django):
     trace_utils.unwrap(django.apps.registry.Apps, "populate")
     trace_utils.unwrap(django.core.handlers.base.BaseHandler, "load_middleware")
     trace_utils.unwrap(django.core.handlers.base.BaseHandler, "get_response")
+    trace_utils.unwrap(django.core.handlers.base.BaseHandler, "get_response_async")
     trace_utils.unwrap(django.template.base.Template, "render")
     trace_utils.unwrap(django.conf.urls.static, "static")
     trace_utils.unwrap(django.conf.urls, "url")
